import {
  ago,
  humanTimedelta,
  preciseHumanTimedelta,
  secondsToTimedelta,
  timedeltaToSeconds
} from '@/helpers'
import { assert } from 'chai'
import sinon from 'sinon'

describe('helpers/date', () => {
  describe('humanTimedelta', () => {
    const CASES = [
      [0, '0 seconds'],
      [0.5, '0 seconds'],
      [1, '1 second'],
      [2, '2 seconds'],
      [59, '59 seconds'],
      [59.999999, '59 seconds'],
      [60, '1 minute'],
      [61, '1 minute'],
      [3600, '1 hour'],
      [3600 * 23, '23 hours'],
      [86400, '1 day'],
      [86400 * 10, '10 days'],
      [86400 * 30, '30 days'],
      [86400 * 3000, '3000 days']
    ]
    for (const [input, expected] of CASES) {
      it(`outputs ${input} seconds as ${expected}`, () => {
        assert.strictEqual(humanTimedelta(input), expected)
      })
    }
  })

  describe('preciseHumanTimedelta', () => {
    const CASES = [
      [0, '0 seconds'],
      [0.000001, '0 seconds'],
      [0.001, '1 millisecond'],
      [0.5, '500 milliseconds'],
      [1, '1 second'],
      [1.001, '1 second 1 millisecond'],
      [59, '59 seconds'],
      [59.999500, '59 seconds 999 milliseconds'],
      [59.999501, '1 minute'],
      [59.999999, '1 minute'],
      [60, '1 minute'],
      [60.001, '1 minute 1 millisecond'],
      [61.001, '1 minute 1 second 1 millisecond'],
      [3600, '1 hour'],
      [3600.001, '1 hour 1 millisecond'],
      [3661.001, '1 hour 1 minute 1 second 1 millisecond'],
      [86399.999, '23 hours 59 minutes 59 seconds 999 milliseconds'],
      [86400, '1 day'],
      [86400 + 3661.001, '1 day 1 hour 1 minute 1 second 1 millisecond'],
      [86400 * 30, '30 days'],
      [86400 * 3000, '3000 days'],
      [86400 * 3000 - 0.001, '2999 days 23 hours 59 minutes 59 seconds 999 milliseconds']
    ]
    for (const [input, expected] of CASES) {
      it(`outputs ${input} seconds as ${expected}`, () => {
        assert.strictEqual(preciseHumanTimedelta(input), expected)
      })
    }
  })

  describe('ago', () => {
    before(() => {
      sinon.stub(Date, 'now').returns(new Date('2020-01-01T00:00:00Z').getTime())
    })
    after(() => {
      sinon.restore()
    })

    const CASES = [
      [0, '0 seconds ago'],
      [0.5, '0 seconds ago'],
      [1, '1 second ago'],
      [2, '2 seconds ago'],
      [59, '59 seconds ago'],
      [59.999999, '1 minute ago'],
      [60, '1 minute ago'],
      [61, '1 minute ago'],
      [3600, '1 hour ago'],
      [3600 * 23, '23 hours ago'],
      [86400, '1 day ago'],
      [86400 * 10, '10 days ago'],
      [86400 * 30, '30 days ago'],
      [86400 * 30 + 1, '2019-12-01 23:59:59'],
      [86400 * 3000, '2011-10-15 00:00:00']
    ]

    for (const [diff, expected] of CASES) {
      it(`outputs ${diff} seconds ago as ${expected}`, () => {
        const input = new Date(Date.now() - diff * 1000)
        assert.strictEqual(ago(input), expected)
      })
    }
  })

  const CONVERSION_CASES = [
    [0, '0:00:00.000000'],
    [1, '0:00:01.000000'],
    [60, '0:01:00.000000'],
    [3600, '1:00:00.000000'],
    [86400, '1 day, 0:00:00.000000'],
    [864000, '10 days, 0:00:00.000000'],
    [86400 + 3600 + 60 + 1.000001, '1 day, 1:01:01.000001'],
    [86400 * 99 + 23 * 3600 + 59 * 60 + 59.999999, '99 days, 23:59:59.999999']
  ]

  describe('secondsToTimedelta', () => {
    for (const [seconds, timedelta] of CONVERSION_CASES) {
      it(`converts ${seconds} seconds to ${timedelta}`, () => {
        assert.strictEqual(secondsToTimedelta(seconds), timedelta)
      })
    }
  })

  describe('timedeltaToSeconds', () => {
    for (const [seconds, timedelta] of CONVERSION_CASES) {
      it(`converts ${timedelta} to ${seconds} seconds`, () => {
        assert.strictEqual(timedeltaToSeconds(timedelta), seconds)
      })
    }
  })
})
