import { assert } from 'chai'
import axios from 'axios'
import { initialState, mutations, getters } from '@/store/annotation.js'
import store from './index.spec.js'
import { FakeAxios, assertRejects } from '../testhelpers.js'

describe('annotation', () => {
  describe('mutations', () => {
    describe('setTextOrientation', () => {
      it('sets the current text orientation', () => {
        const state = initialState()
        assert.strictEqual(state.textOrientation, 'horizontal-lr')
        mutations.setTextOrientation(state, 'horizontal-rl')
        assert.strictEqual(state.textOrientation, 'horizontal-rl')
        mutations.setTextOrientation(state, 'horizontal-lr')
        assert.strictEqual(state.textOrientation, 'horizontal-lr')
      })

      it('forbids unknown text orientation', () => {
        const state = initialState()
        assert.throws(
          () => mutations.setTextOrientation(state, 'boom'),
          Error,
          'Unknown text orientation boom'
        )
      })
    })

    describe('toggle', () => {
      it('toggles without a value', () => {
        const state = initialState()
        assert.ok(!state.enabled)
        mutations.toggle(state)
        assert.ok(state.enabled)
      })

      it('sets directly with a boolean value', () => {
        const state = initialState()
        assert.ok(!state.enabled)
        mutations.toggle(state, false)
        assert.ok(!state.enabled)
        mutations.toggle(state, true)
        assert.ok(state.enabled)
      })

      it('resets the selected tool when turned off', () => {
        const state = initialState()
        state.tool = 'rectangle'
        state.enabled = true
        mutations.toggle(state, false)
        assert.ok(!state.enabled)
        assert.strictEqual(state.tool, 'select')
      })

      it('discards an unsaved polygon when turned off', () => {
        const state = initialState()
        state.enabled = true
        state.selectedElement = { id: 'created-polygon' }
        mutations.toggle(state, false)
        assert.ok(!state.enabled)
        assert.strictEqual(state.selectedElement, null)
      })
      it('turns off batch deletion when turned off', () => {
        const state = initialState()
        state.enabled = true
        state.batchDeletion = true
        mutations.toggle(state, false)
        assert.ok(!state.batchDeletion)
      })
    })

    describe('setTool', () => {
      it('sets the current tool', () => {
        const state = initialState()
        assert.strictEqual(state.tool, 'select')
        mutations.setTool(state, 'rectangle')
        assert.strictEqual(state.tool, 'rectangle')
        mutations.setTool(state, 'polygon')
        assert.strictEqual(state.tool, 'polygon')
      })

      it('forbids unknown tool names', () => {
        const state = initialState()
        assert.throws(
          () => mutations.setTool(state, 'squircle'),
          Error,
          'Unknown tool squircle'
        )
      })
    })

    describe('setBatchCreation', () => {
      it('toggles without a value', () => {
        const state = initialState()
        assert.ok(!state.batchCreation)
        mutations.setBatchCreation(state)
        assert.ok(state.batchCreation)
      })

      it('sets directly with a boolean value', () => {
        const state = initialState()
        assert.ok(!state.batchCreation)
        mutations.setBatchCreation(state, false)
        assert.ok(!state.batchCreation)
        mutations.setBatchCreation(state, true)
        assert.ok(state.batchCreation)
      })
    })

    describe('setBatchTypeEdition', () => {
      it('toggles without a value', () => {
        const state = initialState()
        assert.ok(!state.batchTypeEdition)
        mutations.setBatchTypeEdition(state)
        assert.ok(state.batchTypeEdition)
      })

      it('sets directly with a boolean value', () => {
        const state = initialState()
        assert.ok(!state.batchTypeEdition)
        mutations.setBatchTypeEdition(state, false)
        assert.ok(!state.batchTypeEdition)
        mutations.setBatchTypeEdition(state, true)
        assert.ok(state.batchTypeEdition)
      })
    })

    describe('setDefaultType', () => {
      it('sets the default type for a corpus', () => {
        const state = initialState()
        state.defaultType = { corpus1: 'element' }
        mutations.setDefaultType(state, { corpusId: 'corpus2', type: 'page' })
        assert.deepStrictEqual(state.defaultType, {
          corpus1: 'element',
          corpus2: 'page'
        })
      })

      it('requires a corpus ID', () => {
        const state = initialState()
        assert.throws(
          () => mutations.setDefaultType(state, { type: 'page' }),
          Error,
          'A corpus ID and a type are required to set a default type.'
        )
      })

      it('requires a type', () => {
        const state = initialState()
        assert.throws(
          () => mutations.setDefaultType(state, { corpusId: 'corpus2' }),
          Error,
          'A corpus ID and a type are required to set a default type.'
        )
      })
    })

    describe('setDefaultClass', () => {
      it('sets the default class ID for a corpus', () => {
        const state = initialState()
        state.defaultClass = { corpus1: 'class1' }
        mutations.setDefaultClass(state, { corpusId: 'corpus2', classId: 'class2' })
        assert.deepStrictEqual(state.defaultClass, {
          corpus1: 'class1',
          corpus2: 'class2'
        })
      })

      it('requires a corpus ID', () => {
        const state = initialState()
        assert.throws(
          () => mutations.setDefaultClass(state, { classId: 'class2' }),
          Error,
          'A corpus ID is required to set a default class.'
        )
      })

      it('allows setting no class ID', () => {
        const state = initialState()
        state.defaultClass = { corpus1: 'class1' }
        mutations.setDefaultClass(state, { corpusId: 'corpus2', classId: null })
        assert.deepStrictEqual(state.defaultClass, {
          corpus1: 'class1',
          corpus2: null
        })
      })
    })

    describe('setVisible', () => {
      it('can show an element', () => {
        const state = initialState()
        mutations.setVisible(state, { parentId: 'element1', id: 'element2', visible: true })
        assert.deepStrictEqual(state.visible, { element1: ['element2'] })
      })

      it('can hide an element', () => {
        const state = initialState()
        state.visible = { element1: ['child1', 'child2'] }
        mutations.setVisible(state, { parentId: 'element1', id: 'child1', visible: false })
        assert.deepStrictEqual(state.visible, { element1: ['child2'] })
      })
    })

    describe('setVisibleBulk', () => {
      it('makes multiple elements visible at once', () => {
        const state = initialState()
        mutations.setVisibleBulk(state, { parentId: 'element1', ids: ['tr1', 'tr2', 'tr3'], visible: true })
        assert.deepStrictEqual(state.visible, { element1: ['tr1', 'tr2', 'tr3'] })
      })

      it('does not create duplicates', () => {
        const state = initialState()
        state.visible = { element1: ['item1', 'item2'] }
        mutations.setVisibleBulk(state, { parentId: 'element1', ids: ['item1', 'item3'], visible: true })
        assert.deepStrictEqual(state.visible, { element1: ['item1', 'item2', 'item3'] })
      })

      it('hides multiple elements at once', () => {
        const state = initialState()
        state.visible = { element1: ['item1', 'item2', 'item3', 'item4'] }
        mutations.setVisibleBulk(state, { parentId: 'element1', ids: ['item1', 'item3', 'item5'], visible: false })
        assert.deepStrictEqual(state.visible, { element1: ['item2', 'item4'] })
      })
    })

    describe('cleanVisible', () => {
      it('clean visible elements', () => {
        const state = initialState()
        state.visible = {
          element1: ['child1', 'child2'],
          element2: ['child3']
        }
        mutations.cleanVisible(state, 'element1')
        assert.deepStrictEqual(state.visible, { element2: ['child3'] })
      })
    })

    it('reset', () => {
      const state = {
        enabled: true,
        tool: 'rectangle',
        batchCreation: true,
        batchTypeEdition: true,
        defaultType: {
          corpus1: 'element'
        },
        defaultClass: {
          corpus2: 'class2'
        }
      }

      mutations.reset(state)

      assert.deepStrictEqual(state, initialState())
    })
  })

  describe('actions', () => {
    let mock

    before('Setting up mocks', () => {
      mock = new FakeAxios(axios)
    })

    afterEach(() => {
      // Remove any handlers, but leave mocking in place
      mock.reset()
      store.reset()
    })

    after('Removing Axios mock', () => {
      // Remove mocking entirely
      mock.restore()
    })

    describe('create', () => {
      beforeEach(() => {
        // Enable the annotation mode by default because most tests will require that
        store.state.annotation.enabled = true
      })

      it('requires the annotation to be enabled', async () => {
        store.state.annotation.enabled = false

        const error = await assertRejects(async () => store.dispatch('annotation/create', {
          corpus: 'corpusid',
          parent: 'parentid',
          name: 'element',
          type: 'type',
          classId: null,
          polygon: [[0, 0], [10, 10], [20, 20]]
        }))

        assert.strictEqual(error.toString(), 'Error: Cannot create annotations when the annotation mode is disabled.')
        assert.deepStrictEqual(store.history, [
          {
            action: 'annotation/create',
            payload: {
              corpus: 'corpusid',
              parent: 'parentid',
              name: 'element',
              type: 'type',
              classId: null,
              polygon: [[0, 0], [10, 10], [20, 20]]
            }
          }
        ])
      })

      it('creates an element', async () => {
        mock.onPost('/elements/create/').reply(201, { id: 'elementid', name: 'element' })

        await store.dispatch('annotation/create', {
          corpus: 'corpusid',
          parent: 'parentid',
          name: 'element',
          type: 'type',
          classId: null,
          polygon: [[0, 0], [10, 10], [20, 20]]
        })

        assert.deepStrictEqual(store.history, [
          {
            action: 'annotation/create',
            payload: {
              corpus: 'corpusid',
              parent: 'parentid',
              name: 'element',
              type: 'type',
              classId: null,
              polygon: [[0, 0], [10, 10], [20, 20]]
            }
          },
          {
            action: 'elements/create',
            payload: {
              corpus: 'corpusid',
              parent: 'parentid',
              name: 'element',
              type: 'type',
              polygon: [[0, 0], [10, 10], [20, 20]]
            }
          },
          {
            mutation: 'elements/set',
            payload: {
              id: 'elementid',
              name: 'element'
            }
          },
          {
            mutation: 'elements/addChildren',
            payload: {
              element: 'parentid',
              children: [
                {
                  id: 'elementid',
                  name: 'element'
                }
              ]
            }
          },
          {
            mutation: 'elements/updatedElements'
          },
          {
            mutation: 'annotation/setVisible',
            payload: {
              parentId: 'parentid',
              id: 'elementid',
              visible: true
            }
          }
        ])
      })

      it('creates an element with a classification', async () => {
        mock.onPost('/elements/create/').reply(201, { id: 'elementid', name: 'element' })
        mock.onPost('/classifications/').reply(201, { id: 'classification1', ml_class: 'classid' })

        await store.dispatch('annotation/create', {
          corpus: 'corpusid',
          parent: 'parentid',
          name: 'element',
          type: 'type',
          classId: 'classid',
          polygon: [[0, 0], [10, 10], [20, 20]]
        })

        assert.deepStrictEqual(store.history, [
          {
            action: 'annotation/create',
            payload: {
              corpus: 'corpusid',
              parent: 'parentid',
              name: 'element',
              type: 'type',
              classId: 'classid',
              polygon: [[0, 0], [10, 10], [20, 20]]
            }
          },
          {
            action: 'elements/create',
            payload: {
              corpus: 'corpusid',
              parent: 'parentid',
              name: 'element',
              type: 'type',
              polygon: [[0, 0], [10, 10], [20, 20]]
            }
          },
          {
            mutation: 'elements/set',
            payload: {
              id: 'elementid',
              name: 'element'
            }
          },
          {
            mutation: 'elements/addChildren',
            payload: {
              element: 'parentid',
              children: [
                {
                  id: 'elementid',
                  name: 'element'
                }
              ]
            }
          },
          {
            mutation: 'elements/updatedElements'
          },
          {
            mutation: 'annotation/setVisible',
            payload: {
              parentId: 'parentid',
              id: 'elementid',
              visible: true
            }
          },
          {
            action: 'classification/create',
            payload: {
              elementId: 'elementid',
              mlClass: 'classid'
            }
          },
          {
            mutation: 'elements/addClassification',
            payload: {
              elementId: 'elementid',
              value: {
                id: 'classification1',
                ml_class: 'classid'
              }
            }
          }
        ])
      })

      it('handles element creation errors', async () => {
        mock.onPost('/elements/create/').reply(400, { detail: ['Oh no!'] })

        await assertRejects(async () => {
          await store.dispatch('annotation/create', {
            corpus: 'corpusid',
            parent: 'parentid',
            name: 'element',
            type: 'type',
            classId: null,
            polygon: [[0, 0], [10, 10], [20, 20]]
          })
        })

        assert.deepStrictEqual(store.history, [
          {
            action: 'annotation/create',
            payload: {
              corpus: 'corpusid',
              parent: 'parentid',
              name: 'element',
              type: 'type',
              classId: null,
              polygon: [[0, 0], [10, 10], [20, 20]]
            }
          },
          {
            action: 'elements/create',
            payload: {
              corpus: 'corpusid',
              parent: 'parentid',
              name: 'element',
              type: 'type',
              polygon: [[0, 0], [10, 10], [20, 20]]
            }
          },
          {
            mutation: 'notifications/notify',
            payload: {
              type: 'error',
              text: 'Oh no!'
            }
          }
        ])
      })

      it('handles classification creation errors', async () => {
        mock.onPost('/elements/create/').reply(201, { id: 'elementid', name: 'element' })
        mock.onPost('/classifications/').reply(400, { detail: ['Oh no!'] })

        await store.dispatch('annotation/create', {
          corpus: 'corpusid',
          parent: 'parentid',
          name: 'element',
          type: 'type',
          classId: 'classid',
          polygon: [[0, 0], [10, 10], [20, 20]]
        })

        assert.deepStrictEqual(store.history, [
          {
            action: 'annotation/create',
            payload: {
              corpus: 'corpusid',
              parent: 'parentid',
              name: 'element',
              type: 'type',
              classId: 'classid',
              polygon: [[0, 0], [10, 10], [20, 20]]
            }
          },
          {
            action: 'elements/create',
            payload: {
              corpus: 'corpusid',
              parent: 'parentid',
              name: 'element',
              type: 'type',
              polygon: [[0, 0], [10, 10], [20, 20]]
            }
          },
          {
            mutation: 'elements/set',
            payload: {
              id: 'elementid',
              name: 'element'
            }
          },
          {
            mutation: 'elements/addChildren',
            payload: {
              element: 'parentid',
              children: [
                {
                  id: 'elementid',
                  name: 'element'
                }
              ]
            }
          },
          {
            mutation: 'elements/updatedElements'
          },
          {
            mutation: 'annotation/setVisible',
            payload: {
              parentId: 'parentid',
              id: 'elementid',
              visible: true
            }
          },
          {
            action: 'classification/create',
            payload: {
              elementId: 'elementid',
              mlClass: 'classid'
            }
          },
          {
            mutation: 'notifications/notify',
            payload: {
              type: 'error',
              text: 'Oh no!'
            }
          }
        ])
      })

      it('requires a corpus ID', async () => {
        const error = await assertRejects(async () => {
          await store.dispatch('annotation/create', {
            parent: 'parentid',
            name: 'element',
            type: 'type',
            classId: null,
            polygon: [[0, 0], [10, 10], [20, 20]]
          })
        })
        assert.strictEqual(error.toString(), 'Error: The corpus ID, parent ID and polygon are required to create an annotation.')
      })

      it('requires a parent ID', async () => {
        const error = await assertRejects(async () => {
          await store.dispatch('annotation/create', {
            corpus: 'corpusid',
            name: 'element',
            type: 'type',
            classId: null,
            polygon: [[0, 0], [10, 10], [20, 20]]
          })
        })
        assert.strictEqual(error.toString(), 'Error: The corpus ID, parent ID and polygon are required to create an annotation.')
      })

      it('requires a polygon', async () => {
        const error = await assertRejects(async () => {
          await store.dispatch('annotation/create', {
            corpus: 'corpusid',
            parent: 'parentid',
            name: 'element',
            type: 'type',
            classId: null
          })
        })
        assert.strictEqual(error.toString(), 'Error: The corpus ID, parent ID and polygon are required to create an annotation.')
      })

      it('requires a name without batch mode', async () => {
        const error = await assertRejects(async () => {
          await store.dispatch('annotation/create', {
            corpus: 'corpusid',
            parent: 'parentid',
            type: 'type',
            classId: null,
            polygon: [[0, 0], [10, 10], [20, 20]]
          })
        })
        assert.strictEqual(error.toString(), 'Error: The name, type and classId are required when not annotating in batch.')
      })

      it('requires a type without batch mode', async () => {
        const error = await assertRejects(async () => {
          await store.dispatch('annotation/create', {
            corpus: 'corpusid',
            parent: 'parentid',
            name: 'element',
            classId: null,
            polygon: [[0, 0], [10, 10], [20, 20]]
          })
        })
        assert.strictEqual(error.toString(), 'Error: The name, type and classId are required when not annotating in batch.')
      })

      it('requires a class ID without batch mode', async () => {
        const error = await assertRejects(async () => {
          await store.dispatch('annotation/create', {
            corpus: 'corpusid',
            parent: 'parentid',
            name: 'element',
            type: 'type',
            polygon: [[0, 0], [10, 10], [20, 20]]
          })
        })
        assert.strictEqual(error.toString(), 'Error: The name, type and classId are required when not annotating in batch.')
      })

      it('uses the default type in batch mode', async () => {
        store.state.annotation.batchCreation = true
        store.state.annotation.defaultType = {
          corpusid: 'type'
        }
        mock.onPost('/elements/create/').reply(201, { id: 'elementid', name: 'element' })

        await store.dispatch('annotation/create', {
          corpus: 'corpusid',
          parent: 'parentid',
          name: 'element',
          classId: null,
          polygon: [[0, 0], [10, 10], [20, 20]]
        })

        assert.deepStrictEqual(store.history, [
          {
            action: 'annotation/create',
            payload: {
              corpus: 'corpusid',
              parent: 'parentid',
              name: 'element',
              classId: null,
              polygon: [[0, 0], [10, 10], [20, 20]]
            }
          },
          {
            mutation: 'annotation/selectElement',
            payload: null
          },
          {
            action: 'elements/create',
            payload: {
              corpus: 'corpusid',
              parent: 'parentid',
              name: 'element',
              type: 'type',
              polygon: [[0, 0], [10, 10], [20, 20]]
            }
          },
          {
            mutation: 'elements/set',
            payload: {
              id: 'elementid',
              name: 'element'
            }
          },
          {
            mutation: 'elements/addChildren',
            payload: {
              element: 'parentid',
              children: [
                {
                  id: 'elementid',
                  name: 'element'
                }
              ]
            }
          },
          {
            mutation: 'elements/updatedElements'
          },
          {
            mutation: 'annotation/setVisible',
            payload: {
              parentId: 'parentid',
              id: 'elementid',
              visible: true
            }
          }
        ])
      })

      it('requires a default type in batch mode', async () => {
        store.state.annotation.batchCreation = true
        const error = await assertRejects(async () => {
          await store.dispatch('annotation/create', {
            corpus: 'corpusid',
            parent: 'parentid',
            name: 'element',
            classId: null,
            polygon: [[0, 0], [10, 10], [20, 20]]
          })
        })
        assert.strictEqual(error.toString(), 'Error: No default type has been set for batch annotation on corpus corpusid.')
      })

      it('uses the default class in batch mode', async () => {
        store.state.annotation.batchCreation = true
        store.state.annotation.defaultClass = {
          corpusid: 'classid'
        }
        mock.onPost('/elements/create/').reply(201, { id: 'elementid', name: 'element' })
        mock.onPost('/classifications/').reply(201, { id: 'classification1', ml_class: 'classid' })

        await store.dispatch('annotation/create', {
          corpus: 'corpusid',
          parent: 'parentid',
          name: 'element',
          type: 'type',
          polygon: [[0, 0], [10, 10], [20, 20]]
        })

        assert.deepStrictEqual(store.history, [
          {
            action: 'annotation/create',
            payload: {
              corpus: 'corpusid',
              parent: 'parentid',
              name: 'element',
              type: 'type',
              polygon: [[0, 0], [10, 10], [20, 20]]
            }
          },
          {
            mutation: 'annotation/selectElement',
            payload: null
          },
          {
            action: 'elements/create',
            payload: {
              corpus: 'corpusid',
              parent: 'parentid',
              name: 'element',
              type: 'type',
              polygon: [[0, 0], [10, 10], [20, 20]]
            }
          },
          {
            mutation: 'elements/set',
            payload: {
              id: 'elementid',
              name: 'element'
            }
          },
          {
            mutation: 'elements/addChildren',
            payload: {
              element: 'parentid',
              children: [
                {
                  id: 'elementid',
                  name: 'element'
                }
              ]
            }
          },
          {
            mutation: 'elements/updatedElements'
          },
          {
            mutation: 'annotation/setVisible',
            payload: {
              parentId: 'parentid',
              id: 'elementid',
              visible: true
            }
          },
          {
            action: 'classification/create',
            payload: {
              elementId: 'elementid',
              mlClass: 'classid'
            }
          },
          {
            mutation: 'elements/addClassification',
            payload: {
              elementId: 'elementid',
              value: {
                id: 'classification1',
                ml_class: 'classid'
              }
            }
          }
        ])
      })

      it('uses the default name in batch mode', async () => {
        store.state.annotation.batchCreation = true
        mock.onPost('/elements/create/').reply(201, { id: 'elementid', name: '1' })

        await store.dispatch('annotation/create', {
          corpus: 'corpusid',
          parent: 'parentid',
          type: 'type',
          classId: null,
          polygon: [[0, 0], [10, 10], [20, 20]]
        })

        assert.deepStrictEqual(store.history, [
          {
            action: 'annotation/create',
            payload: {
              corpus: 'corpusid',
              parent: 'parentid',
              type: 'type',
              classId: null,
              polygon: [[0, 0], [10, 10], [20, 20]]
            }
          },
          {
            mutation: 'annotation/selectElement',
            payload: null
          },
          {
            action: 'elements/create',
            payload: {
              corpus: 'corpusid',
              parent: 'parentid',
              name: '1',
              type: 'type',
              polygon: [[0, 0], [10, 10], [20, 20]]
            }
          },
          {
            mutation: 'elements/set',
            payload: {
              id: 'elementid',
              name: '1'
            }
          },
          {
            mutation: 'elements/addChildren',
            payload: {
              element: 'parentid',
              children: [
                {
                  id: 'elementid',
                  name: '1'
                }
              ]
            }
          },
          {
            mutation: 'elements/updatedElements'
          },
          {
            mutation: 'annotation/setVisible',
            payload: {
              parentId: 'parentid',
              id: 'elementid',
              visible: true
            }
          }
        ])
      })
    })

    describe('typeEdit', () => {
      beforeEach(() => {
        // Enable the annotation mode by default because most tests will require that
        store.state.annotation.enabled = true
      })

      it('requires the annotation to be enabled', async () => {
        store.state.annotation.enabled = false

        const element = {
          id: 'elementId',
          corpus: {
            id: 'corpusId',
            name: 'Some Corpus'
          },
          type: 'holotype'
        }

        const error = await assertRejects(async () => store.dispatch('annotation/typeEdit', { element }))

        assert.strictEqual(error.toString(), 'Error: Cannot edit element types when the annotation mode is disabled.')
        assert.deepStrictEqual(store.history, [
          {
            action: 'annotation/typeEdit',
            payload: {
              element
            }
          }
        ])
      })

      it('requires a type to be set', async () => {
        const element = {
          id: 'elementId',
          corpus: {
            id: 'corpusId',
            name: 'Some Corpus'
          },
          type: 'holotype'
        }

        const error = await assertRejects(async () => store.dispatch('annotation/typeEdit', {
          element
        }))

        assert.strictEqual(error.toString(), "Error: A valid type value is required to update the element's type.")
        assert.deepStrictEqual(store.history, [
          {
            action: 'annotation/typeEdit',
            payload: {
              element
            }
          }
        ])
      })

      it("updates the element's type", async () => {
        const element = {
          id: 'elementId',
          corpus: {
            id: 'corpusId',
            name: 'Some Corpus'
          },
          type: 'holotype'
        }
        const updatedElement = {
          id: 'elementId',
          name: 'elementName',
          type: 'cyanotype',
          zone: {
            polygon: [[0, 0], [10, 10], [20, 20]]
          }
        }
        mock.onPatch('/element/elementId/').reply(201, updatedElement)

        await store.dispatch('annotation/typeEdit', {
          element, type: 'cyanotype'
        })

        assert.deepStrictEqual(store.history, [
          {
            action: 'annotation/typeEdit',
            payload: {
              element,
              type: 'cyanotype'
            }
          },
          {
            action: 'elements/patch',
            payload: {
              id: 'elementId',
              type: 'cyanotype'
            }
          },
          {
            mutation: 'elements/set',
            payload: updatedElement
          },
          {
            mutation: 'elements/updatedElements'
          }
        ])
      })

      it('requires a default type to be set in batch mode', async () => {
        store.state.annotation.defaultType = { wrong_corpus: 'some type' }
        store.state.annotation.batchTypeEdition = true

        const element = {
          id: 'elementId',
          corpus: {
            id: 'corpusId',
            name: 'Some Corpus'
          },
          type: 'holotype'
        }

        const error = await assertRejects(async () => store.dispatch('annotation/typeEdit', {
          element
        }))

        assert.strictEqual(error.toString(), "Error: A valid type value is required to update the element's type.")
        assert.deepStrictEqual(store.history, [
          {
            action: 'annotation/typeEdit',
            payload: {
              element
            }
          }
        ])
      })

      it('updates the target element\'s type with the default type in batch mode', async () => {
        store.state.annotation.defaultType = { corpusId: 'cyanotype' }
        store.state.annotation.batchTypeEdition = true
        const element = {
          id: 'elementId',
          name: 'elementName',
          type: 'phenotype',
          corpus: {
            id: 'corpusId',
            name: 'Some Corpus'
          },
          zone: {
            polygon: [[0, 0], [10, 10], [20, 20]]
          }
        }
        store.state.elements = {
          elements: {
            elementId: element
          }
        }
        const updatedElement = {
          id: 'elementId',
          name: 'elementName',
          type: 'cyanotype',
          zone: {
            polygon: [[0, 0], [10, 10], [20, 20]]
          }
        }
        mock.onPatch('/element/elementId/').reply(201, updatedElement)

        await store.dispatch('annotation/typeEdit', {
          element
        })

        assert.deepStrictEqual(store.history, [
          {
            action: 'annotation/typeEdit',
            payload: {
              element: {
                id: 'elementId',
                name: 'elementName',
                type: 'phenotype',
                corpus: {
                  id: 'corpusId',
                  name: 'Some Corpus'
                },
                zone: {
                  polygon: [[0, 0], [10, 10], [20, 20]]
                }
              }
            }
          },
          {
            action: 'elements/patch',
            payload: {
              id: 'elementId',
              type: 'cyanotype'
            }
          },
          {
            mutation: 'elements/set',
            payload: updatedElement
          },
          {
            mutation: 'elements/updatedElements'
          }
        ])
      })

      it("does not send an API request if the selected type is already the element's type in batch mode", async () => {
        store.state.annotation.defaultType = { corpusId: 'cyanotype' }
        store.state.annotation.batchTypeEdition = true
        const element = {
          id: 'elementId',
          name: 'elementName',
          type: 'cyanotype',
          corpus: {
            id: 'corpusId',
            name: 'Some Corpus'
          },
          zone: {
            polygon: [[0, 0], [10, 10], [20, 20]]
          }
        }
        await store.dispatch('annotation/typeEdit', {
          element
        })

        assert.deepStrictEqual(store.history, [
          {
            action: 'annotation/typeEdit',
            payload: {
              element: {
                id: 'elementId',
                name: 'elementName',
                type: 'cyanotype',
                corpus: {
                  id: 'corpusId',
                  name: 'Some Corpus'
                },
                zone: {
                  polygon: [[0, 0], [10, 10], [20, 20]]
                }
              }
            }
          }
        ])
      })

      it('handles errors in batch mode', async () => {
        store.state.annotation.defaultType = { corpusId: 'cyanotype' }
        store.state.annotation.batchTypeEdition = true
        const element = {
          id: 'elementId',
          name: 'elementName',
          type: 'phenotype',
          corpus: {
            id: 'corpusId',
            name: 'Some Corpus'
          },
          zone: {
            polygon: [[0, 0], [10, 10], [20, 20]]
          }
        }
        mock.onPatch('/element/elementId/').reply(400, { detail: ['Oh no!'] })
        await assertRejects(async () => {
          await store.dispatch('annotation/typeEdit', {
            element
          })
        })

        assert.deepStrictEqual(store.history, [
          {
            action: 'annotation/typeEdit',
            payload: {
              element: {
                id: 'elementId',
                name: 'elementName',
                type: 'phenotype',
                corpus: {
                  id: 'corpusId',
                  name: 'Some Corpus'
                },
                zone: {
                  polygon: [[0, 0], [10, 10], [20, 20]]
                }
              }
            }
          },
          {
            action: 'elements/patch',
            payload: {
              id: 'elementId',
              type: 'cyanotype'
            }
          },
          {
            mutation: 'notifications/notify',
            payload: {
              text: 'Oh no!',
              type: 'error'
            }
          }
        ])
      })
    })
  })

  describe('getters', () => {
    describe('defaultName', () => {
      it('generates a default name', () => {
        const rootState = {
          elements: {
            links: {
              parentid: {
                children: ['child1', 'child2']
              }
            },
            elements: {
              child1: {
                id: 'child1',
                type: 'page'
              },
              child2: {
                id: 'child2',
                type: 'page'
              }
            }
          }
        }

        assert.strictEqual(
          getters.defaultName({}, {}, rootState)('parentid', 'page'),
          '3'
        )
      })

      it('handles a parent having no children', () => {
        const rootState = {
          elements: {
            links: {
              parentid: {
                children: []
              }
            }
          }
        }

        assert.strictEqual(
          getters.defaultName({}, {}, rootState)('parentid', 'page'),
          '1'
        )
      })

      it('handles a parent with no defined links', () => {
        const rootState = {
          elements: {
            links: {}
          }
        }

        assert.strictEqual(
          getters.defaultName({}, {}, rootState)('parentid', 'page'),
          '1'
        )
      })

      it('handles links to unknown elements', () => {
        const rootState = {
          elements: {
            links: {
              parentid: {
                children: [
                  'lostchild',
                  'child2',
                  'child1'
                ]
              }
            },
            elements: {
              child1: {
                id: 'child1',
                type: 'page'
              },
              child2: {
                id: 'child2',
                type: 'page'
              }
            }
          }
        }

        assert.strictEqual(
          getters.defaultName({}, {}, rootState)('parentid', 'page'),
          // The third child is ignored as we can't tell its type
          '3'
        )
      })

      it('only counts elements of the correct type', () => {
        const rootState = {
          elements: {
            links: {
              parentid: {
                children: ['child1', 'child2']
              }
            },
            elements: {
              child1: {
                id: 'child1',
                type: 'page'
              },
              child2: {
                id: 'child2',
                type: 'act'
              }
            }
          }
        }

        assert.strictEqual(
          getters.defaultName({}, {}, rootState)('parentid', 'page'),
          '2'
        )
      })
    })
  })
})
