import { assert } from 'chai'
import axios from 'axios'
import { initialState, mutations } from '@/store/repos.js'
import store from './index.spec.js'
import { useWorkerStore } from '@/stores'
import { createPinia, setActivePinia } from 'pinia'
import { repoSample, reposSample } from '../samples.js'
import { FakeAxios } from '../testhelpers.js'

describe('repos', () => {
  describe('mutations', () => {
    it('setRepos', () => {
      const state = initialState()
      mutations.setRepos(state, [
        repoSample,
        { id: 'second_repo', name: 'repo2' }
      ])
      assert.deepStrictEqual(state.repositories, {
        repoid: repoSample,
        second_repo: { id: 'second_repo', name: 'repo2' }
      })
    })

    it('removeRepo', () => {
      const state = {
        repositories: { repoid: repoSample }
      }
      mutations.removeRepo(state, 'repoid')
      assert.deepStrictEqual(state.repositories, {})
    })

    it('reset', () => {
      const state = {
        repositories: { repoid: repoSample }
      }
      mutations.reset(state)
      assert.deepStrictEqual(state, initialState())
    })
  })

  describe('actions', () => {
    let mock, workerStore

    before('Setting up Axios mock', () => {
      mock = new FakeAxios(axios)
      setActivePinia(createPinia())
      workerStore = useWorkerStore()
    })

    afterEach(() => {
      // Remove any handlers, but leave mocking in place
      mock.reset()
      store.reset()
      workerStore.$reset()
    })

    after('Removing Axios mock', () => {
      // Remove mocking entirely
      mock.restore()
    })

    describe('list', () => {
      it('lists existing repositories', async () => {
        const sample = {
          ...reposSample,
          count: 2,
          results: [
            ...reposSample.results,
            {
              id: 'other_repo',
              workers: [
                {
                  id: 'worker 2'
                }, {
                  id: 'worker 3'
                }
              ]
            }
          ]
        }
        mock.onGet('/process/repos/').reply(200, sample)

        const page = await store.dispatch('repos/list', {})

        assert.deepStrictEqual(store.history, [
          { action: 'repos/list', payload: {} },
          { mutation: 'repos/setRepos', payload: sample.results }
        ])
        assert.deepStrictEqual(page, sample)
        assert.deepStrictEqual(Object.keys(workerStore.workers), ['worker 1', 'worker 2', 'worker 3'])
      })
    })

    describe('delete', () => {
      it('deletes a repo', async () => {
        mock.onDelete('/process/repos/repoid/').reply(204)

        store.state.repositories = { repoid: {}, secondId: {} }
        await store.dispatch('repos/delete', 'repoid')

        assert.deepStrictEqual(store.history, [
          {
            action: 'repos/delete',
            payload: 'repoid'
          },
          {
            mutation: 'repos/removeRepo',
            payload: 'repoid'
          }
        ])
      })
    })
  })
})
