import { assert } from 'chai'
import axios from 'axios'
import { FakeAxios, assertRejects, setUpTestPinia, actionsCompletedPlugin } from '../testhelpers'
import { useDatasetStore, useElementsStore, useNotificationStore } from '@/stores'

describe('datasets', () => {
  describe('actions', () => {
    let mock, store, notificationStore, elementsStore

    before('Setting up mocks', () => {
      mock = new FakeAxios(axios)
      setUpTestPinia([actionsCompletedPlugin])
      store = useDatasetStore()
      notificationStore = useNotificationStore()
      elementsStore = useElementsStore()
    })

    afterEach(() => {
      // Remove any handlers, but leave mocking in place
      mock.reset()
      store.$reset()
      notificationStore.$reset()
      elementsStore.$reset()
    })

    after('Removing Axios mock', () => {
      // Remove mocking entirely
      mock.restore()
    })

    describe('listCorpusDatasets', () => {
      it('list datasets in a corpus', async () => {
        const pages = [
          {
            count: 3,
            number: 1,
            next: 'nextpage',
            results: [
              {
                id: 'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa',
                state: 'open',
                name: 'Shin Seiki Evangelion',
                description: 'Become legend, young boy!',
                sets: ['unit-00', 'unit-01', 'unit-02']
              },
              {
                id: 'bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb',
                state: 'open',
                name: 'Mobile Suit Gundam Wing',
                description: 'another dataset',
                sets: ['training', 'test', 'validation']
              }
            ]
          },
          {
            count: 3,
            number: 2,
            next: null,
            results: [
              {
                id: 'cccccccc-cccc-cccc-cccc-cccccccccccc',
                state: 'open',
                name: 'Sailor Moon',
                description: 'Moon prism power, make up!',
                sets: ['neptune', 'uranus', 'pluto', 'saturn']
              }
            ]
          }
        ]
        mock.onGet('/corpus/corpus1/datasets/', { params: { page: 1 } }).reply(200, pages[0])
        mock.onGet('/corpus/corpus1/datasets/', { params: { page: 2 } }).reply(200, pages[1])

        await store.listCorpusDatasets('corpus1')

        assert.deepStrictEqual(store.corpusDatasets, {
          corpus1: new Set([
            'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa',
            'bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb',
            'cccccccc-cccc-cccc-cccc-cccccccccccc'
          ])
        })

        assert.deepStrictEqual(store.datasets, {
          'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa': {
            id: 'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa',
            state: 'open',
            name: 'Shin Seiki Evangelion',
            description: 'Become legend, young boy!',
            sets: ['unit-00', 'unit-01', 'unit-02']
          },
          'bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb': {
            id: 'bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb',
            state: 'open',
            name: 'Mobile Suit Gundam Wing',
            description: 'another dataset',
            sets: ['training', 'test', 'validation']
          },
          'cccccccc-cccc-cccc-cccc-cccccccccccc': {
            id: 'cccccccc-cccc-cccc-cccc-cccccccccccc',
            state: 'open',
            name: 'Sailor Moon',
            description: 'Moon prism power, make up!',
            sets: ['neptune', 'uranus', 'pluto', 'saturn']
          }
        })
      })
    })

    describe('createCorpusDataset', () => {
      it('creates a new dataset in a corpus', async () => {
        store.corpusDatasets = {
          corpus1: new Set([
            'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa',
            'bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb'
          ])
        }
        store.datasets = {
          'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa': {
            id: 'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa',
            state: 'open',
            name: 'Shin Seiki Evangelion',
            description: 'Become legend, young boy!',
            sets: ['unit-00', 'unit-01', 'unit-02']
          },
          'bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb': {
            id: 'bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb',
            state: 'open',
            name: 'Mobile Suit Gundam Wing',
            description: 'another dataset',
            sets: ['training', 'test', 'validation']
          }
        }
        const requestData = {
          name: 'Sailor Moon',
          description: 'Moon prism power, make up!',
          sets: ['neptune', 'uranus', 'pluto', 'saturn']
        }
        const responseData = {
          ...requestData,
          id: 'cccccccc-cccc-cccc-cccc-cccccccccccc',
          state: 'open'
        }
        mock.onPost('/corpus/corpus1/datasets/', { ...requestData }).reply(201, { ...responseData })

        await store.createCorpusDataset('corpus1', requestData)

        assert.deepStrictEqual(store.datasets, {
          'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa': {
            id: 'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa',
            state: 'open',
            name: 'Shin Seiki Evangelion',
            description: 'Become legend, young boy!',
            sets: ['unit-00', 'unit-01', 'unit-02']
          },
          'bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb': {
            id: 'bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb',
            state: 'open',
            name: 'Mobile Suit Gundam Wing',
            description: 'another dataset',
            sets: ['training', 'test', 'validation']
          },
          'cccccccc-cccc-cccc-cccc-cccccccccccc': {
            id: 'cccccccc-cccc-cccc-cccc-cccccccccccc',
            state: 'open',
            name: 'Sailor Moon',
            description: 'Moon prism power, make up!',
            sets: ['neptune', 'uranus', 'pluto', 'saturn']
          }
        })
        assert.deepStrictEqual(store.corpusDatasets, {
          corpus1: new Set([
            'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa',
            'bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb',
            'cccccccc-cccc-cccc-cccc-cccccccccccc'
          ])
        })
      })

      it('handles errors', async () => {
        const requestData = {
          name: 'Sailor Moon',
          description: 'Moon prism power, make up!',
          sets: ['neptune', 'uranus', 'pluto', 'saturn']
        }
        mock.onPost('/corpus/corpus1/datasets/', { ...requestData }).reply(400, { detail: ['no'] })
        await assertRejects(async () => store.createCorpusDataset('corpus1', requestData))
        assert.deepStrictEqual(notificationStore.notifications, [
          {
            id: 0,
            type: 'error',
            text: 'no'
          }
        ])
      })
    })

    describe('updateCorpusDataset', () => {
      it('updates an existing dataset in a corpus', async () => {
        store.datasets = {
          'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa': {
            id: 'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa',
            state: 'open',
            name: 'Shin Seiki Evangelion',
            description: 'Become legend, young boy!',
            sets: ['unit-00', 'unit-01', 'unit-02']
          },
          'bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb': {
            id: 'bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb',
            state: 'open',
            name: 'Mobile Suit Gundam Wing',
            description: 'another dataset',
            sets: ['training', 'test', 'validation']
          }
        }
        store.corpusDatasets = {
          corpus1: new Set([
            'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa',
            'bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb'
          ])
        }
        const responseData = {
          id: 'bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb',
          state: 'open',
          name: 'Pretty Guardian Sailor Moon',
          description: 'another dataset',
          sets: ['training', 'test', 'validation']
        }
        mock.onPatch('/datasets/bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb/', { name: 'Pretty Guardian Sailor Moon' }).reply(200, { ...responseData })

        await store.updateCorpusDataset(
          {
            id: 'bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb',
            name: 'Pretty Guardian Sailor Moon'
          }
        )

        assert.deepStrictEqual(store.datasets, {
          'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa': {
            id: 'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa',
            state: 'open',
            name: 'Shin Seiki Evangelion',
            description: 'Become legend, young boy!',
            sets: ['unit-00', 'unit-01', 'unit-02']
          },
          'bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb': {
            id: 'bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb',
            state: 'open',
            name: 'Pretty Guardian Sailor Moon',
            description: 'another dataset',
            sets: ['training', 'test', 'validation']
          }
        })
        assert.deepStrictEqual(store.corpusDatasets, {
          corpus1: new Set([
            'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa',
            'bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb'
          ])
        })
      })

      it('handles errors', async () => {
        mock.onPatch('/datasets/bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb/', { name: 'Pretty Guardian Sailor Moon' }).reply(400, { detail: ['no'] })
        await assertRejects(async () => store.updateCorpusDataset(
          {
            id: 'bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb',
            name: 'Pretty Guardian Sailor Moon'
          }
        ))
        assert.deepStrictEqual(notificationStore.notifications, [
          {
            id: 0,
            type: 'error',
            text: 'no'
          }
        ])
      })
    })

    describe('deleteCorpusDataset', () => {
      it('deletes an existing dataset', async () => {
        store.datasets = {
          'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa': {
            id: 'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa',
            state: 'open',
            name: 'Shin Seiki Evangelion',
            description: 'Become legend, young boy!',
            sets: ['unit-00', 'unit-01', 'unit-02']
          },
          'bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb': {
            id: 'bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb',
            state: 'open',
            name: 'Mobile Suit Gundam Wing',
            description: 'another dataset',
            sets: ['training', 'test', 'validation']
          }
        }
        store.corpusDatasets = {
          corpus1: new Set([
            'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa',
            'bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb'
          ])
        }
        mock.onDelete('/datasets/bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb/').reply(204)

        await store.deleteCorpusDataset(
          'corpus1',
          'bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb'
        )

        assert.deepStrictEqual(store.datasets, {
          'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa': {
            id: 'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa',
            state: 'open',
            name: 'Shin Seiki Evangelion',
            description: 'Become legend, young boy!',
            sets: ['unit-00', 'unit-01', 'unit-02']
          }
        })
        assert.deepStrictEqual(store.corpusDatasets, {
          corpus1: new Set([
            'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa'
          ])
        })
      })

      it('handles errors', async () => {
        mock.onDelete('/datasets/bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb/').reply(400, { detail: ['no'] })
        await assertRejects(async () => store.deleteCorpusDataset(
          'corpus1',
          'bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb'
        ))
        assert.deepStrictEqual(notificationStore.notifications, [
          {
            id: 0,
            type: 'error',
            text: 'no'
          }
        ])
      })
    })

    describe('addDatasetElementsSelection', () => {
      it('adds elements from a selection to a dataset', async () => {
        mock.onPost('/corpus/corpusid/datasets/selection/', { dataset_id: 'datasetid', set: 'test' }).reply(204)

        const params = { dataset_id: 'datasetid', set: 'test' }
        await store.addDatasetElementsSelection('corpusid', params)

        assert.deepStrictEqual(notificationStore.notifications, [
          {
            id: 0,
            type: 'success',
            text: 'Elements have been added to the dataset.'
          }
        ])
      })

      it('automatically resets dataset elements pagination', async () => {
        store.datasetElementPagination = {
          datasetid: {
            test: {
              datasetElements: [{ set: 'test', element: { id: 'firstElement' } }],
              nextCursor: '1234'
            },
            test2: {
              datasetElements: [{ set: 'test', element: { id: 'firstElement' } }],
              nextCursor: '5678'
            }
          }
        }

        const params = { dataset_id: 'datasetid', set: 'test' }
        mock.onPost('/corpus/corpusid/datasets/selection/', params).reply(204)
        await store.addDatasetElementsSelection('corpusid', params)

        assert.deepStrictEqual(store.datasetElementPagination, {
          datasetid: {
            test2: {
              datasetElements: [{ set: 'test', element: { id: 'firstElement' } }],
              nextCursor: '5678'
            }
          }
        })
      })

      it('throw errors', async () => {
        mock.onPost('/corpus/corpusid/datasets/selection/', { dataset_id: 'datasetid', set: 'test' }).reply(400)

        const params = { dataset_id: 'datasetid', set: 'test' }
        await assertRejects(
          async () => await store.addDatasetElementsSelection('corpusid', params)
        )

        assert.deepStrictEqual(notificationStore.notifications, [
          {
            id: 0,
            type: 'error',
            text: 'Request failed with status code 400'
          }
        ])
      })
    })

    describe('singleCorpusDatasets', () => {
      it('returns the datasets from a given corpus', async () => {
        store.datasets = {
          'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa': {
            id: 'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa',
            state: 'open',
            name: 'Shin Seiki Evangelion',
            description: 'Become legend, young boy!',
            sets: ['unit-00', 'unit-01', 'unit-02']
          },
          'bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb': {
            id: 'bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb',
            state: 'open',
            name: 'Mobile Suit Gundam Wing',
            description: 'another dataset',
            sets: ['training', 'test', 'validation']
          },
          'cccccccc-cccc-cccc-cccc-cccccccccccc': {
            id: 'cccccccc-cccc-cccc-cccc-cccccccccccc',
            state: 'open',
            name: 'Sailor Moon',
            description: 'Moon prism power, make up!',
            sets: ['neptune', 'uranus', 'pluto', 'saturn']
          }
        }
        store.corpusDatasets = {
          corpus1: new Set(['aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa']),
          corpus2: new Set([
            'bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb',
            'cccccccc-cccc-cccc-cccc-cccccccccccc'
          ])
        }

        assert.deepStrictEqual(store.singleCorpusDatasets('corpus2'), [
          {
            id: 'bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb',
            state: 'open',
            name: 'Mobile Suit Gundam Wing',
            description: 'another dataset',
            sets: ['training', 'test', 'validation']
          },
          {
            id: 'cccccccc-cccc-cccc-cccc-cccccccccccc',
            state: 'open',
            name: 'Sailor Moon',
            description: 'Moon prism power, make up!',
            sets: ['neptune', 'uranus', 'pluto', 'saturn']
          }
        ])
      })

      it('handles a non-existent corpus ID', async () => {
        store.datasets = {
          'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa': {
            id: 'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa',
            state: 'open',
            name: 'Shin Seiki Evangelion',
            description: 'Become legend, young boy!',
            sets: ['unit-00', 'unit-01', 'unit-02']
          }
        }
        store.corpusDatasets = {
          corpus1: new Set(['aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa'])
        }

        try {
          store.singleCorpusDatasets('corpus2')
        } catch (err) {
          assert.equal(err.message, 'Datasets for corpus corpus2 could not be retrieved.')
        }
      })
    })

    describe('nextDatasetElements', () => {
      it('Lists first page elements from the set of a dataset', async () => {
        const page = {
          count: 3,
          number: null,
          next: 'https://arkindex.teklia.com/api/v1/datasets/aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa/elements/?cursor=1234',
          previous: null,
          results: [
            {
              set: 'unit-00',
              element: {
                id: '1',
                type: 'page',
                name: 'asuka',
                corpus: {
                  id: 'corpus1'
                },
                thumbnail_url: null,
                zone: null,
                rotation_angle: 0,
                mirrored: false,
                classes: [],
                metadata: 0,
                has_children: false,
                created: 'somedate'
              }
            },
            {
              set: 'unit-00',
              element: {
                id: '2',
                type: 'page',
                name: 'rei',
                corpus: {
                  id: 'corpus1'
                },
                thumbnail_url: null,
                zone: null,
                rotation_angle: 0,
                mirrored: false,
                classes: [],
                metadata: 0,
                has_children: false,
                created: 'somedate'
              }
            }
          ]
        }

        mock.onGet('/datasets/aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa/elements/', { params: { cursor: null, set: 'unit-00' } }).reply(200, page)

        await store.nextDatasetElements('aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa', 'unit-00')
        await store.actionsCompleted()
        assert.deepStrictEqual(store.datasetElementPagination, {
          'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa': {
            'unit-00': {
              datasetElements: page.results,
              nextCursor: '1234'
            }
          }
        })
      })

      it('Iteratively lists elements from the set of a dataset', async () => {
        const nextPage = {
          count: 2,
          number: null,
          next: null,
          previous: 'https://arkindex.teklia.com/api/v1/datasets/aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa/elements/?cursor=0789',
          results: [
            {
              set: 'unit-00',
              element: {
                id: '3',
                type: 'page',
                name: 'shinji',
                corpus: {
                  id: 'corpus1'
                },
                thumbnail_url: null,
                zone: null,
                rotation_angle: 0,
                mirrored: false,
                classes: [],
                metadata: 0,
                has_children: false,
                created: 'somedate'
              }
            }
          ]
        }
        store.datasetElementPagination = {
          'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa': {
            'unit-00': {
              datasetElements: [{ set: 'unit-00', element: { id: 'firstElement' } }],
              nextCursor: '1234'
            }
          }
        }

        mock.onGet('/datasets/aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa/elements/', { params: { cursor: '1234', set: 'unit-00' } }).reply(200, nextPage)

        await store.nextDatasetElements('aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa', 'unit-00')
        assert.deepStrictEqual(store.datasetElementPagination, {
          'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa': {
            'unit-00': {
              datasetElements: [
                { set: 'unit-00', element: { id: 'firstElement' } },
                ...nextPage.results
              ],
              nextCursor: null
            }
          }
        })

        // No next page
        await store.nextDatasetElements('aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa', 'unit-00')
        assert.deepStrictEqual(notificationStore.notifications, [
          {
            id: 0,
            type: 'warning',
            text: 'All elements have been fetched already.'
          }
        ])
      })

      it('handles errors', async () => {
        mock.onGet('/datasets/aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa/elements/', { params: { cursor: null, set: 'unit-00' } }).reply(400, { detail: ['no'] })

        await assertRejects(
          async () => await store.nextDatasetElements('aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa', 'unit-00')
        )

        assert.deepStrictEqual(store.datasetElementPagination, { 'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa': {} })
        assert.deepStrictEqual(notificationStore.notifications, [
          {
            id: 0,
            type: 'error',
            text: 'no'
          }
        ])
      })
    })

    describe('removeDatasetElement', () => {
      it('removes an element from a dataset', async () => {
        const elements = [
          {
            set: 'unit-00',
            element: {
              id: '1',
              type: 'page',
              name: 'asuka',
              corpus: {
                id: 'corpus1'
              },
              thumbnail_url: null,
              zone: null,
              rotation_angle: 0,
              mirrored: false,
              classes: [],
              metadata: 0,
              has_children: false,
              created: 'somedate'
            }
          },
          {
            set: 'unit-00',
            element: {
              id: '2',
              type: 'page',
              name: 'shinji',
              corpus: {
                id: 'corpus1'
              },
              thumbnail_url: null,
              zone: null,
              rotation_angle: 0,
              mirrored: false,
              classes: [],
              metadata: 0,
              has_children: false,
              created: 'somedate'
            }
          }
        ]
        store.datasetElementPagination = {
          'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa': {
            'unit-00': {
              datasetElements: elements,
              nextCursor: '1234'
            }
          }
        }
        elementsStore.elementDatasets = {
          2: [
            {
              set: 'unit-00',
              dataset: {
                id: 'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa',
                state: 'open',
                name: 'Shin Seiki Evangelion',
                description: 'Become legend, young boy!',
                sets: ['unit-00', 'unit-01', 'unit-02']
              }
            },
            {
              set: 'unit-01',
              dataset: {
                id: 'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa',
                state: 'open',
                name: 'Set 01',
                description: 'A',
                sets: ['unit-00', 'unit-01', 'unit-02']
              }
            }
          ]
        }

        mock.onDelete('/datasets/aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa/elements/2/', 'unit-00').reply(204)

        await store.removeDatasetElement('aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa', '2', 'unit-00')
        await store.actionsCompleted()

        assert.deepStrictEqual(store.datasetElementPagination, {
          'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa': {
            'unit-00': {
              datasetElements: [elements[0]],
              nextCursor: '1234'
            }
          }
        })
        assert.deepStrictEqual(elementsStore.elementDatasets, {
          2: [
            {
              set: 'unit-01',
              dataset: {
                id: 'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa',
                state: 'open',
                name: 'Set 01',
                description: 'A',
                sets: ['unit-00', 'unit-01', 'unit-02']
              }
            }
          ]
        })
      })

      it('handles errors', async () => {
        const elements = [
          {
            set: 'unit-00',
            element: {
              id: '1',
              type: 'page',
              name: 'asuka',
              corpus: {
                id: 'corpus1'
              },
              thumbnail_url: null,
              zone: null,
              rotation_angle: 0,
              mirrored: false,
              classes: [],
              metadata: 0,
              has_children: false,
              created: 'somedate'
            }
          },
          {
            set: 'unit-01',
            element: {
              id: '2',
              type: 'page',
              name: 'rei',
              corpus: {
                id: 'corpus1'
              },
              thumbnail_url: null,
              zone: null,
              rotation_angle: 0,
              mirrored: false,
              classes: [],
              metadata: 0,
              has_children: false,
              created: 'somedate'
            }
          },
          {
            set: 'unit-00',
            element: {
              id: '3',
              type: 'page',
              name: 'shinji',
              corpus: {
                id: 'corpus1'
              },
              thumbnail_url: null,
              zone: null,
              rotation_angle: 0,
              mirrored: false,
              classes: [],
              metadata: 0,
              has_children: false,
              created: 'somedate'
            }
          }
        ]
        store.datasetElements = {
          'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa': elements
        }

        mock.onDelete('/datasets/aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa/elements/2/', 'unit-01').reply(400, { detail: ['no'] })

        await assertRejects(
          async () => await store.removeDatasetElement('aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa', '2', 'unit-01')
        )

        assert.deepStrictEqual(store.datasetElements, {
          'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa': elements
        })
        assert.deepStrictEqual(notificationStore.notifications, [
          {
            id: 0,
            type: 'error',
            text: 'no'
          }
        ])
      })
    })

    describe('cloneDataset', () => {
      it('clone an existing dataset', async () => {
        store.corpusDatasets = {
          corpus1: new Set(['aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa'])
        }
        store.datasets = {
          'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa': {
            id: 'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa',
            state: 'error',
            name: 'A',
            corpus_id: 'corpus1'
          }
        }
        mock.onPost('/datasets/aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa/clone/').reply(201, {
          id: 'bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb',
          state: 'open',
          name: 'Clone of A',
          corpus_id: 'corpus1'
        })

        await store.cloneDataset('aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa')

        assert.deepStrictEqual(store.datasets, {
          'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa': {
            id: 'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa',
            state: 'error',
            name: 'A',
            corpus_id: 'corpus1'
          },
          'bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb': {
            id: 'bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb',
            state: 'open',
            name: 'Clone of A',
            corpus_id: 'corpus1'
          }
        })
        assert.deepStrictEqual(store.corpusDatasets, {
          corpus1: new Set([
            'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa',
            'bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb'
          ])
        })
      })

      it('handles errors', async () => {
        mock.onPost('/datasets/aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa/clone/').reply(400, { detail: ['no'] })
        await assertRejects(async () => store.cloneDataset('aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa'))
        assert.deepStrictEqual(notificationStore.notifications, [
          {
            id: 0,
            type: 'error',
            text: 'no'
          }
        ])
      })
    })
  })
})
