import { assert } from 'chai'
import axios from 'axios'
import { FakeAxios, setUpTestPinia, actionsCompletedPlugin } from '../testhelpers'
import { useTranscriptionStore, useNotificationStore } from '@/stores'
import { transcriptionsPage1, transcriptionsPage2 } from '../samples.js'

describe('transcription', () => {
  describe('actions', () => {
    let mock, store, notificationStore

    before('Setting up mocks', () => {
      mock = new FakeAxios(axios)
      setUpTestPinia([actionsCompletedPlugin])
      store = useTranscriptionStore()
      notificationStore = useNotificationStore()
    })

    afterEach(() => {
      // Remove any handlers, but leave mocking in place
      mock.reset()
      store.$reset()
      notificationStore.$reset()
    })

    after('Removing Axios mock', () => {
      // Remove mocking entirely
      mock.restore()
    })

    describe('list', () => {
      it("lists an element's transcriptions", async () => {
        mock.onGet('/element/elementid/transcriptions/', { params: { page: 1 } }).reply(200, transcriptionsPage1)
        mock.onGet('/element/elementid/transcriptions/', { params: { page: 2 } }).reply(200, transcriptionsPage2)
        await store.list('elementid')
        await store.actionsCompleted()

        assert.deepStrictEqual(store.transcriptions, {
          elementid: Object.fromEntries(
            [...transcriptionsPage1.results, ...transcriptionsPage2.results].map(t => [t.id, t])
          )
        })
      })
    })

    describe('create', () => {
      it('creates a new transcription on an element', async () => {
        const transcription = {
          id: 'transcriptionid',
          text: 'Steelix'
        }
        mock.onPost('/element/elementid/transcription/').reply(201, transcription)
        await store.create('elementid', { text: 'Steelix' })
        assert.deepStrictEqual(store.transcriptions, {
          elementid: {
            transcriptionid: transcription
          }
        })
      })
    })

    describe('update', () => {
      it('updateTranscription', async () => {
        store.transcriptions = {
          elementid: {
            transcriptionid: {
              id: 'transcriptionid',
              text: 'Steelix'
            }
          }
        }
        mock.onPatch('/transcription/transcriptionid/').reply(200, { id: 'transcriptionid', text: 'Animula vagula blandula' })
        const updated = await store.update('elementid', 'transcriptionid', { text: 'Animula vagula blandula' })
        assert.deepStrictEqual(store.transcriptions, {
          elementid: {
            transcriptionid: updated
          }
        })
      })
    })

    describe('delete', () => {
      it("deletes an element's transcription", async () => {
        store.transcriptions = {
          elementid: {
            transcriptionid: {
              id: 'transcriptionid',
              text: 'Steelix'
            }
          }
        }
        mock.onDelete('/transcription/transcriptionid/').reply(204)
        await store.delete('elementid', 'transcriptionid')
        assert.deepStrictEqual(store.transcriptions, { elementid: {} })
      })
    })
  })
})
