/* eslint-disable @typescript-eslint/no-var-requires */
const { defineConfig } = require('@vue/cli-service')
const CompressionPlugin = require('compression-webpack-plugin')
const { LicenseWebpackPlugin } = require('license-webpack-plugin')
const LodashModuleReplacementPlugin = require('lodash-webpack-plugin')
const MiniCssExtractPlugin = require('mini-css-extract-plugin')

const { version } = require('./package.json')

process.env.VUE_APP_VERSION = process.env.NODE_ENV === 'development' ? `${version} (local dev)` : version

module.exports = defineConfig({
  lintOnSave: false,
  chainWebpack: config => {
    config.module.merge({
      parser: {
        javascript: {
          exportsPresence: 'error'
        }
      }
    })

    config
      .plugin('lodash')
      .use(LodashModuleReplacementPlugin, [{
        // See https://github.com/lodash/lodash-webpack-plugin#feature-sets
        shorthands: true,
        cloning: false,
        currying: true,
        caching: true,
        collections: true,
        exotics: true,
        guards: true,
        metadata: true,
        deburring: true,
        unicode: true,
        chaining: true,
        memoizing: true,
        coercions: true,
        flattening: true,
        paths: false,
        placeholders: true
      }])

    config
      .stats({
        children: false,
        modules: false,
        chunks: false
      })

    config
      .when(process.env.NODE_ENV === 'production', config => {
        config
          .output
          .filename(`arkindex-${version}.js`)
          .chunkFilename(`arkindex-[name]-${version}.js`)

        config
          .plugin('compression')
          .use(CompressionPlugin)

        config
          .plugin('license')
          .use(LicenseWebpackPlugin, [{
            outputFilename: 'licenses-' + version + '.txt',
            perChunkOutput: false,
            addBanner: true
          }])

        // Override the config of the existing MiniCssExtractPlugin to set our own filenames
        config
          .plugin('extract-css')
          .use(MiniCssExtractPlugin, [{
            filename: `arkindex-${version}.css`,
            chunkFilename: `arkindex-[name]-${version}.css`
          }])

        /*
         * On Surge, we need to use the content hash to cause a browser cache invalidation
         * when someone visits Surge twice on the same URL, but with different builds.
         */
        config
          .when(process.env.TARGET === 'surge', config => {
            config
              .output
              .filename('arkindex-[contenthash].js')
              .chunkFilename('arkindex-[name]-[contenthash].js')

            // Override the config of the existing MiniCssExtractPlugin to set our own filenames
            config
              .plugin('extract-css')
              .use(MiniCssExtractPlugin, [{
                filename: 'arkindex-[contenthash].css',
                chunkFilename: 'arkindex-[name]-[contenthash].css'
              }])
          })

        config
          .when(process.env.TARGET !== 'surge', config => {
            config.optimization.merge({
              splitChunks: {
                cacheGroups: {
                  defaultVendors: {
                    /*
                     * The vendors chunk uses `filename` instead of `chunkFilename`, which causes duplicate names
                     * between the entrypoint and the vendors chunk, so we override the name here.
                     * This override is not necessary on Surge builds since they use the content hash.
                     */
                    filename: `arkindex-vendors-${version}.js`
                  }
                }
              }
            })
          })
      })

    config
      .when(process.env.NODE_ENV !== 'production', config => {
        config
          .output
          .filename('arkindex.js')
          .chunkFilename('arkindex-[name].js')

        config.optimization.merge({
          splitChunks: {
            cacheGroups: {
              defaultVendors: {
                /*
                 * The vendors chunk uses `filename` instead of `chunkFilename`, which causes duplicate names
                 * between the entrypoint and the vendors chunk, so we override the name here.
                 */
                filename: 'arkindex-vendors.js'
              }
            }
          }
        })

        config.devtool('source-map')
      })
  }
})
